/*
 * libxml-dtc provides an XML framework creation facilities
 * Copyright (C) 2003-2004 Mike Hommey <mh@glandium.org>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 */

#define _GNU_SOURCE

#include <modules.h>
#include <xml-dtc.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <dirent.h>
#include <config.h>

#ifdef MODULES_DIR
#include <ltdl.h>
#endif

#include "static.h"

struct _xml_dtc xml_dtc = {
  NULL, NULL,
  NULL, NULL, NULL
};

int xml_dtc_cleanup() {
  if (xml_dtc.hooks)
    xml_dtc_list_free(xml_dtc.hooks);
  if (xml_dtc.callers)
    xml_dtc_hash_free(xml_dtc.callers);
  if (xml_dtc.components)
    xml_dtc_hash_free(xml_dtc.components);
  if (xml_dtc.modules) {
    int i;
    for (i = 0; i < xml_dtc.modules->num; i++) {
      if (((xml_dtc_module *)xml_dtc.modules->list[i].element)->close)
	((xml_dtc_module *)xml_dtc.modules->list[i].element)->close();
#ifdef MODULES_DIR
      if (((xml_dtc_module *)xml_dtc.modules->list[i].element)->_private)
	lt_dlclose(((xml_dtc_module *)xml_dtc.modules->list[i].element)->_private);
#endif
    }
    xml_dtc_hash_free(xml_dtc.modules);
  }
  if (xml_dtc.resource)
    xml_dtc_resource_unlink(xml_dtc.resource);
  memset(&xml_dtc, 0, sizeof(xml_dtc));
  return 0;
}

#ifdef MODULES_DIR
static int xml_dtc_load_module(char *filename) {
  lt_dlhandle dl;
  xml_dtc_module *module;

  xml_dtc_msg(xml_dtc.resource, XML_DTC_MSG_DEBUG, "Loading %s module.", filename);

  /* TODO: Handle errors */
  dl = lt_dlopen(filename);
  if (!dl) {
    xml_dtc_msg(xml_dtc.resource, XML_DTC_MSG_ERROR, "%s",lt_dlerror());
    return -1;
  } else {
    const char *error;
    char *module_name, *tmp;

    module_name = rindex(filename, '/');
    if (module_name) {
      module_name++;
    } else {
      module_name = filename;
    }
    module_name[strlen(module_name) - 3] = 0;
    tmp = malloc(16 + strlen(module_name));
    sprintf(tmp, "xml_dtc_%s_module", module_name);
    module_name[strlen(module_name) - 3] = '.';
    module = lt_dlsym(dl, tmp);
    free(tmp);
    if ((error = lt_dlerror()) != NULL) {
      xml_dtc_msg(xml_dtc.resource, XML_DTC_MSG_ERROR, "%s",error);
      return -1;
    } else {
      if (module->API == XML_DTC_API_VERSION) {
	if (xml_dtc_register_module(module) == 0)
	  module->_private = dl;
      } else {
	xml_dtc_msg(xml_dtc.resource, XML_DTC_MSG_ERROR, "API version mismatch.");
	return -1;
      }
    }
  }

  return 0;
}

static int xml_dtc_recursively_load_modules(char *path) {
  DIR *dir;
  struct dirent *current;

  xml_dtc_msg(xml_dtc.resource, XML_DTC_MSG_DEBUG, "Loading modules from %s.", path);

  dir = opendir(path);

  if (dir) {
    while ((current = readdir(dir))) {
      if (current->d_type == DT_DIR) {
	if (! ((current->d_name[0] == '.') && ((current->d_name[1] == 0) || ((current->d_name[1] == '.') && (current->d_name[2] == 0))))) {
	  char *new_dir;
	  
	  new_dir = malloc(strlen(path) + strlen(current->d_name) + 2);
	  strncpy(new_dir, path, strlen(path));
	  new_dir[strlen(path)] = '/';
	  strncpy(new_dir + strlen(path) + 1, current->d_name, strlen(current->d_name) + 1);
	  
	  xml_dtc_recursively_load_modules(new_dir);
	  
	  free(new_dir);	  
	}
      } else if (current->d_type == DT_REG) {
	int len;
	if (((len = strlen(current->d_name)) > 3) && (current->d_name[len - 1] == 'a') && (current->d_name[len - 2] == 'l') && (current->d_name[len - 3] == '.')) {
	  char *new_dir;
	  
	  new_dir = malloc(strlen(path) + strlen(current->d_name) + 2);
	  strncpy(new_dir, path, strlen(path));
	  new_dir[strlen(path)] = '/';
	  strncpy(new_dir + strlen(path) + 1, current->d_name, strlen(current->d_name) + 1);
	  
	  xml_dtc_load_module(new_dir);

	  free(new_dir);	  
	}
      }
    }
    closedir(dir);
    return 0;
  }
  return -1;
}
#endif

int xml_dtc_register_module(xml_dtc_module *module) {
  if (xml_dtc_hash_add(xml_dtc.modules, (char *)module->name, module) == 0) {
    xml_dtc_hash *hash;
    xml_dtc_component_def *component = (xml_dtc_component_def *) module->components;
    xml_dtc_caller_def *caller = (xml_dtc_caller_def *) module->callers;

    module->registr();

    if (caller)
      while(caller->name) {
	if (! xml_dtc.callers)
	  xml_dtc.callers = xml_dtc_hash_new(xml_dtc.resource, NULL, NULL, NULL, NULL);
	
	if (xml_dtc_hash_add(xml_dtc.callers, (char *)caller->name, (char *)caller)) {
	  xml_dtc_msg(xml_dtc.resource, XML_DTC_MSG_ERROR, "Cannot register caller '%s': already registered.", caller->name);
	} else {
	  if (caller->component) {
	    xml_dtc_msg(xml_dtc.resource, XML_DTC_MSG_DEBUG, "Registering caller '%s' to type '%s'.", caller->name, caller->component);
	  } else {
	    xml_dtc_msg(xml_dtc.resource, XML_DTC_MSG_DEBUG, "Registering '%s' caller.", caller->name);
	  }
	  if (! xml_dtc.components) 
	    xml_dtc.components = xml_dtc_hash_new_cast(xml_dtc.resource, NULL, NULL, xml_dtc_hash_free, NULL);
	  hash = xml_dtc_hash_new(xml_dtc.resource, NULL, NULL, NULL, NULL);
	  xml_dtc_hash_add(xml_dtc.components, (char *)caller->component, hash);
	}
	
	caller++;
      }
    if (xml_dtc.components && component)
      while(component->name) {
	if ((hash = xml_dtc_hash_get(xml_dtc.components, component->type))) {
	  xml_dtc_hash_add(hash, (char *)component->name, component);
	  xml_dtc_msg(xml_dtc.resource, XML_DTC_MSG_DEBUG, "Registering '%s' %s.", component->name, component->type);
	}
	component++;
      }
    return 0;
  } else {
    xml_dtc_msg(xml_dtc.resource, XML_DTC_MSG_ERROR, "Module \"%s\" already loaded.", module->name);
    return -1;
  }
}

int xml_dtc_init_modules() {
  int i = 0;
  xml_dtc_module *module;

  while ((module = xml_dtc_static_modules[i])) {
    xml_dtc_msg(xml_dtc.resource, XML_DTC_MSG_DEBUG, "Initializing internal module \"%s\"", module->name);

    if (xml_dtc.modules == NULL)
      xml_dtc.modules = xml_dtc_hash_new(xml_dtc.resource, NULL, NULL, NULL, NULL);

    if (xml_dtc_register_module(module) == 0)
      module->_private = NULL;
    i++;
  }

#ifdef MODULES_DIR
  if (lt_dlinit() == 0) {
    char *path;
    path = getenv("XML_DTC_MODULES_DIR");
    if (path == NULL)
      path = MODULES_DIR;

    xml_dtc_recursively_load_modules(path);
  }
#endif

  if (!xml_dtc.modules) {
    xml_dtc_msg(xml_dtc.resource, XML_DTC_MSG_ERROR, "Cannot initialize " PACKAGE ".");
    xml_dtc_cleanup();
    return -1;
  }
  return 0;
}

int xml_dtc_init(const xml_dtc_resource *resource) {
  if (xml_dtc.resource)
    return -1;
  
  if (! resource)
    resource = xml_dtc_resource_new();
  
  //  xml_dtc_resource_set_msg_verbosity((xml_dtc_resource *)resource, XML_DTC_MSG_DEBUG);
  
  xml_dtc.resource = xml_dtc_resource_link(resource);

  return 0;
}

int xml_dtc_register_hook(xml_dtc_hook callback) {
  xml_dtc_msg(xml_dtc.resource, XML_DTC_MSG_DEBUG, "Registering a hook.");
  if (! xml_dtc.hooks) {
    xml_dtc.hooks = xml_dtc_value_list_new(xml_dtc.resource);
  }
  xml_dtc_list_push_cast(xml_dtc.hooks, callback);

  return 0;
}
