/*
 * This module provides actions to the XML-DTC framework
 * Copyright (C) 2003-2004 Mike Hommey <mh@glandium.org>
 * 
 * This module is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This module is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this module; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 */

#include <modules.h>
#include <xml-dtc.h>
#include <mappings.h>
#include <list.h>
#include <string.h>

static xml_dtc_component *xml_dtc_handler_action_callback(xml_dtc_handler *handler, xml_dtc_hash *handlers, xml_dtc_sitemap *sitemap, xmlNodePtr node) {
  xmlChar *attr;
  xml_dtc_component *result;

  attr = xmlGetProp(node, "name");
  result = xml_dtc_component_new(sitemap->resource, attr, "action");
  result->handle = handler->_private;
  free(attr);

  xml_dtc_sitemap_add(sitemap, result);
  return (xml_dtc_component *)handler;
}

static xml_dtc_caller *xml_dtc_handler_act_callback(xml_dtc_handler *handler, xml_dtc_hash *handlers, xml_dtc_sitemap *sitemap, xmlNodePtr node) {
  xml_dtc_caller *result;

  result = xml_dtc_caller_new(sitemap->resource, "action", "action");
  result->component.name = xmlGetProp(node, "type");

  node = node->children;

  while (node) {
    /* Ignore Blank Nodes */
    if (! xmlIsBlankNode(node)) {
      /* Test if node is an element or something else (PI, Comment...) */
      if (node->type != XML_ELEMENT_NODE) {
        /* Comments are allowed */
        if (node->type == XML_COMMENT_NODE) {
          xml_dtc_msg(sitemap->resource, XML_DTC_MSG_DEBUG, "Found comment");
        } else {
          xml_dtc_msg(sitemap->resource, XML_DTC_MSG_DEBUG, "Element or comment expected");
          return NULL;
        }
      } else {
        /* Check if namespace is good */
        if ((node->ns != node->parent->ns) && (xmlStrcmp(node->ns->href, XML_DTC_SITEMAP_NAMESPACE) != 0)) {
          xml_dtc_msg(sitemap->resource, XML_DTC_MSG_ERROR, "expecting %s namespace, but got %s", XML_DTC_SITEMAP_NAMESPACE,
                      (node->ns != NULL) ? (char *) node->ns->href : "none");
          return NULL;
        }
        /* Now we know that namespace is ok, we can check and initialize all stuff */
        if (xmlStrcmp(node->name, "parameter") == 0) {
          xmlChar *tmp1, *tmp2;
          if (!result->params) {
            result->params = xml_dtc_string_hash_new(result->resource);
          }
          xml_dtc_string_hash_add(result->params, tmp1 = xmlGetProp(node, "name"
), tmp2 = xmlGetProp(node, "value"));
          xmlFree(tmp1);
          xmlFree(tmp2);
        } else {
          break;
        }
      }
    }
    node = node->next;
  }

  result->_private = xml_dtc_sitemap_get_pipeline(sitemap, node);
  result->free_private = (int (*) (void *))xml_dtc_list_free;

  return result;
}

static int xml_dtc_act_http_header(xml_dtc_caller *caller, xml_dtc_request *request) {
  int i;
  char *key, *value;

  if (caller->params) {
    for(i = 0; i < caller->params->num; i++) {
      key = xml_dtc_resolve_mappings(request->mappings, caller->params->list[i].key);
      value = xml_dtc_resolve_mappings(request->mappings, (char *)caller->params->list[i].element);
      
      xml_dtc_request_set_out_header(request, key, value);
      free(value);
      free(key);
    }
  }
  xml_dtc_mappings_new_list(request->mappings);
  i = xml_dtc_request_apply(request, caller->_private);
  xml_dtc_mappings_destroy_last(request->mappings);
  return i;
}

static int xml_dtc_act_set_cookie(xml_dtc_caller *caller, xml_dtc_request *request) {
  int i;
  char *key, *value;

  if (caller->params) {
    for(i = 0; i < caller->params->num; i++) {
      key = xml_dtc_resolve_mappings(request->mappings, caller->params->list[i].key);
      value = xml_dtc_resolve_mappings(request->mappings, (char *)caller->params->list[i].element);
      
      xml_dtc_request_set_cookie(request, key, value);
      free(value);
      free(key);
    }
  }
  xml_dtc_mappings_new_list(request->mappings);
  i = xml_dtc_request_apply(request, caller->_private);
  xml_dtc_mappings_destroy_last(request->mappings);
  return i;
}

static int xml_dtc_actions_module_register(void) {
  return 0;
}

static const xml_dtc_caller_def xml_dtc_actions_callers[] = {
  { "act", "action", (xml_dtc_handler_callback) xml_dtc_handler_act_callback },
  { NULL }
};

static const xml_dtc_component_def xml_dtc_actions[] = {
  { "http_header", "action", xml_dtc_act_http_header, (xml_dtc_handler_callback) xml_dtc_handler_action_callback },
  { "set_cookie", "action", xml_dtc_act_set_cookie, (xml_dtc_handler_callback) xml_dtc_handler_action_callback },
  { NULL }
};

xml_dtc_module xml_dtc_actions_module = {
  XML_DTC_MODULE_STANDARD_STUFF,
  xml_dtc_actions_module_register,
  NULL,
  xml_dtc_actions_callers,
  xml_dtc_actions
};
