/*
 * libxml-dtc provides an XML framework creation facilities
 * Copyright (C) 2003-2004 Mike Hommey <mh@glandium.org>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 */

#include <string.h>
#include <components.h>

int _xml_dtc_caller_free(xml_dtc_caller *caller) {
  if (caller) {
    if (caller->resolved == 0)
      free(caller->component.name);
    else
      xml_dtc_component_unlink(caller->component.object);
    if (caller->params)
      xml_dtc_hash_free(caller->params);
  }
  return 0;
}

xml_dtc_component *xml_dtc_component_new(const xml_dtc_resource *resource, const char *name, const char *type) {
  xml_dtc_component *result;
  char *the_type = xml_dtc_hash_get_key(xml_dtc.components, type);

  if (! the_type) {
    xml_dtc_msg(resource, XML_DTC_MSG_ERROR, "Attempt to create non existent component type '%s'.", type);
    return NULL;
  }
  if (! name) {
    xml_dtc_msg(resource, XML_DTC_MSG_ERROR, "Attempt to create a component without a name.");
    return NULL;
  }

  result = (xml_dtc_component *) malloc(sizeof(xml_dtc_component));
  memset(result, 0, sizeof(xml_dtc_component));

  result->resource = xml_dtc_resource_link(resource);

  result->name = strdup(name);

  result->type = the_type;

  return result;
}

xml_dtc_component *xml_dtc_component_link(xml_dtc_component *component) {
  if (component)
    component->links++;
  return component;
}

int xml_dtc_component_unlink(xml_dtc_component *component) {
  if (component) {
    if (component->links-- > 1) {
      return 1;
    } else {
      if (component->_private) {
	if (component->free_private)
	  component->free_private(component->_private);
	else
	  free(component->_private);
      }
      xml_dtc_resource_unlink(component->resource);
      xml_dtc_list_free(component->config);
      free(component->name);
      free(component);
      return 0;
    }
  }
  return -1;
}

xml_dtc_caller *xml_dtc_caller_new(const xml_dtc_resource *resource, const char *name, const char *type) {
  xml_dtc_caller *result;
  char *the_type = xml_dtc_hash_get(xml_dtc.components, type);

  if (! the_type) {
    xml_dtc_msg(resource, XML_DTC_MSG_ERROR, "Attempt to create non existent caller type '%s'.", type);
    return NULL;
  }
  if (! name) {
    xml_dtc_msg(resource, XML_DTC_MSG_ERROR, "Attempt to create a caller without a name.");
    return NULL;
  }

  result = malloc(sizeof(xml_dtc_caller));
  memset(result, 0, sizeof(xml_dtc_caller));

  result->resource = xml_dtc_resource_link(resource);

  result->name = strdup(name);

  result->type = the_type;

  return result;
}

int xml_dtc_caller_unlink(xml_dtc_caller *caller) {
  if (caller) {
    if (caller->links-- > 1) {
      return 1;
    } else {
      if (caller->_private) {
	if (caller->free_private)
	  caller->free_private(caller->_private);
	else
	  free(caller->_private);
      }
      xml_dtc_resource_unlink(caller->resource);
      switch (caller->resolved) {
      case 1:
        xml_dtc_component_unlink(caller->component.object);
        break;
      case 2:
        break;
      default:
        xmlFree(caller->component.name);
      }
      if (caller->params)
        xml_dtc_hash_free(caller->params);
      free(caller->name);
      free(caller);
      return 0;
    }
  }
  return -1;
}
