/*
 * libxml-dtc provides an XML framework creation facilities
 * Copyright (C) 2003-2004 Mike Hommey <mh@glandium.org>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 */

#define _GNU_SOURCE

#include <request.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <dlfcn.h>
#include <unistd.h>

int xml_dtc_request_free(xml_dtc_request *request) {
  if (request) {
    if (request->error.description)
      free(request->error.description);
    if (request->error.type)
      free(request->error.type);
    if (request->error.message)
      free(request->error.message);
    if (request->mappings)
      xml_dtc_mappings_free(request->mappings);
    if (request->xmltree)
      xmlFreeDoc(request->xmltree);
    if (request->result)
      xmlFree(request->result);
    if (request->method)
      free(request->method);
    if (request->uri)
      free(request->uri);
    if (request->path_info)
      free(request->path_info);
    if (request->attributes)
      xml_dtc_hash_free(request->attributes);
    if (request->cookies)
      xml_dtc_hash_free(request->cookies);
    if (request->newcookies)
      xml_dtc_hash_free(request->newcookies);
    if (request->headers)
      xml_dtc_hash_free(request->headers);
    if (request->parameters)
      xml_dtc_hash_free(request->parameters);
    if (request->resource)
      xml_dtc_resource_unlink(request->resource);
    free(request);
    return 0;
  }
  return -1;
}

int xml_dtc_request_default_register_cookies(xml_dtc_request *request);

xml_dtc_request *xml_dtc_request_new(const xml_dtc_resource *resource, const char *method, const char *uri, const char *path_info) {
  xml_dtc_request *result;
  char *path;

  result = (xml_dtc_request *)malloc(sizeof(xml_dtc_request));
  memset(result, 0, sizeof(xml_dtc_request));
  result->resource = xml_dtc_resource_link(resource);
  result->mappings = xml_dtc_mappings_new(resource);
  if (method)
    result->method = strdup(method);
  if (uri)
    result->uri = strdup(uri);
  if (path_info)
    result->path_info = strdup(path_info);
  path = result->path_info;

  while (result->path_info[0] == '/') {
    result->path_info++;
  }

  if (result->path_info != path) {
    char *tmp;
    tmp = strdup(result->path_info);
    result->path_info = tmp;
    free(path);
  }

  result->status = 200;
  result->register_cookies = xml_dtc_request_default_register_cookies;

  return result;
}

int xml_dtc_request_set_error(xml_dtc_request *request, const xml_dtc_component *sender, const char *type, const char *message, const char *description) {
  if (request) {
    request->error.type = strdup(type);
    request->error.message = strdup(message);
    request->error.description = strdup(description);
    request->error.sender = sender;
    return 0;
  } else {
    return -1;
  }
}

int xml_dtc_request_serialize_error(xml_dtc_request *request) {
  char sender[256];
  if (request->error.sender == (xml_dtc_component *) request->sitemap) {
    strncpy(sender, "sitemap", 8);
  } else {
    snprintf(sender, 255, "%s %s", request->error.sender->name, request->error.sender->type);
  }
  request->result = malloc(4096);
  snprintf(request->result, 4095,
	  "<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>\n" \
	  "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Strict//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd\">\n" \
	  "<html xmlns=\"http://www.w3.org/1999/xhtml\">\n" \
	  "<head>\n" \
	  "<title>Resource Not Found</title>\n" \
	  "<style type=\"text/css\">\n" \
	  "html { font-family: sans-serif; line-height: 1.3em }\n" \
	  "h1 { font-size: 1.3em; border-bottom: 1px solid black; }\n" \
	  "dt { font-weight: bold }\n" \
	  "dd { margin-left: 8em; margin-top: -1.3em; margin-bottom: .5em; }\n" \
	  "</style>\n" \
	  "</head>\n" \
	  "<body>\n" \
	  "<h1>&lt;xml-dtc/&gt; - %s</h1>\n" \
	  "<dl>\n" \
	  "<dt>type</dt>\n" \
	  "<dd>%s</dd>\n" \
	  "<dt>message</dt>\n" \
	  "<dd>%s</dd>\n" \
	  "<dt>description</dt>\n" \
	  "<dd>%s</dd>\n" \
	  "<dt>sender</dt>\n" \
	  "<dd>%s</dd>\n" \
	  "<!-- Add stack trace if necessary -->\n" \
	  "<dt>request-uri</dt>\n" \
	  "<dd>%s</dd>\n" \
	  "<dt>path-info</dt>\n" \
	  "<dd>%s</dd>\n" \
	  "</dl>\n" \
	  "</body>\n" \
	  "</html>\n",
	  request->error.message,
	  request->error.type,
	  request->error.message,
	  request->error.description,
	  sender,
	  request->uri,
	  request->path_info);
  request->content_length = -1;
  request->mime_type = "text/html";

  return 0;
}

xml_dtc_hash *xml_dtc_request_get_parameters(xml_dtc_request *request) {
  if (request) {
    if (request->parameters == NULL) {
      if (request->register_parameters)
	request->register_parameters(request);
      else
	return NULL;
    }
    return request->parameters;
  }
  return NULL;
}

char *xml_dtc_request_get_parameter(xml_dtc_request *request, const char *name) {
  return xml_dtc_hash_get(xml_dtc_request_get_parameters(request), name);
}

int xml_dtc_request_set_parameter(xml_dtc_request *request, const char *name, const char *value) {
  if (request) {
    if (request->parameters == NULL)
      request->parameters = xml_dtc_string_hash_new(request->resource);
    xml_dtc_string_hash_add(request->parameters, name, value);
    return 0;
  }
  return -1;
}

int xml_dtc_request_set_parameters_from_query_string(xml_dtc_request *request, const char *query_string) {
  if (request && query_string) {
    char *query, *amp, *eq, *name, *value, *end;
    query = strdup(query_string);
    amp = query - 1;
    end = index(query, 0);
    while (amp) {
      name = amp + 1;
      value = NULL;
      eq = index(name, '=');
      amp = index(name, '&');
      if ((eq == name) || ((amp == 0) && (eq == 0) && (name >= end)) || (amp == name))
	name = NULL;
      else {
	if (eq) {
	  if ((eq < amp) || (amp == 0)) {
	    *eq = 0;
	    if (amp != eq + 1)
	      value = eq + 1;
	  }
	}
	if (amp)
	  *amp = 0;
      }
      xml_dtc_request_set_parameter(request, name, value);
    }
    free(query);
  }

  return 0;
}


xml_dtc_hash *xml_dtc_request_get_attributes(xml_dtc_request *request) {
  return request->attributes;
}

char *xml_dtc_request_get_attribute(xml_dtc_request *request, const char *name) {
  if (request && request->attributes) {
    return xml_dtc_hash_get(request->attributes, name);
  }
  return NULL;
}

int xml_dtc_request_set_attribute(xml_dtc_request *request, const char *name, const char *value) {
  if (request) {
    if (request->attributes == NULL)
      request->attributes = xml_dtc_string_hash_new(request->resource);
    xml_dtc_string_hash_add(request->attributes, name, value);
    return 0;
  }
  return -1;
}

xml_dtc_hash *xml_dtc_request_get_headers(xml_dtc_request *request) {
  if (request) {
    if (request->headers == NULL) {
      if (request->register_headers)
	request->register_headers(request);
      else
	return NULL;
    }
    return request->headers;
  }
  return NULL;
}

char *xml_dtc_request_get_header(xml_dtc_request *request, const char *name) {
  char *new_name;
  char *result;
  int i;
  new_name = strdup(name);
  for (i = 0; new_name[i]; i++) 
    if ((new_name[i] >= 'A') && (new_name[i] <= 'Z'))
      new_name[i] -= 'A' - 'a';
    else if (new_name[i] == '_')
      new_name[i] = '-';
  result = xml_dtc_hash_get(xml_dtc_request_get_headers(request), new_name);
  free(new_name);
  return result;
}

int xml_dtc_request_set_header(xml_dtc_request *request, const char *name, const char *value) {
  if (request) {
    char *new_name;
    int i;
    if (request->headers == NULL)
      request->headers = xml_dtc_string_hash_new(request->resource);
    new_name = strdup(name);
    for (i = 0; new_name[i]; i++) {
      if ((new_name[i] >= 'A') && (new_name[i] <= 'Z'))
	new_name[i] -= 'A' - 'a';
      else if (new_name[i] == '_')
	new_name[i] = '-';
    }
    xml_dtc_string_hash_add(request->headers, new_name, value);
    free(new_name);
    return 0;
  }
  return -1;
}

xml_dtc_hash *xml_dtc_request_get_cookies(xml_dtc_request *request) {
  if (request) {
    if (request->cookies == NULL) {
      if (request->register_cookies)
	request->register_cookies(request);
      else
	return NULL;
    }
    return request->cookies;
  }
  return NULL;
}

char *xml_dtc_request_get_cookie(xml_dtc_request *request, const char *name) {
  char *result;
  if ((result = xml_dtc_hash_get(request->newcookies, name))) {
    return result;
  } else
    return xml_dtc_hash_get(xml_dtc_request_get_cookies(request), name);
}

int xml_dtc_request_set_cookie(xml_dtc_request *request, const char *name, const char *value) {
  if (request) {
    if (request->newcookies == NULL)
      request->newcookies = xml_dtc_string_hash_new(request->resource);
    xml_dtc_string_hash_add(request->newcookies, name, value);
    return 0;
  }
  return -1;
}

int xml_dtc_request_register_cookie(xml_dtc_request *request, const char *name, const char *value) {
  if (request) {
    if (request->cookies == NULL)
      request->cookies = xml_dtc_string_hash_new(request->resource);
    xml_dtc_string_hash_add(request->cookies, name, value);
    return 0;
  }
  return -1;
}

int xml_dtc_request_default_register_cookies(xml_dtc_request *request) {
  if (request) {
    char *cookies = xml_dtc_request_get_header(request, "Cookie");
    if (cookies) {
      char *cook, *sc, *eq, *name, *value, *end;
      cook = strdup(cookies);
      sc = cook - 1;
      end = index(cook, 0);
      while (sc) {
	name = sc + 1;
	value = NULL;
	eq = index(name, '=');
	sc = index(name, ';');
	if ((eq == name) || ((sc == 0) && (eq == 0) && (name >= end)) || (sc == name))
	  name = NULL;
	else {
	  if (eq) {
	    if ((eq < sc) || (sc == 0)) {
	      *eq = 0;
	      if (sc != eq + 1)
		value = eq + 1;
	    }
	  }
	  if (sc)
	    *sc = 0;
	}
	while (*name == ' ') { name++; }
	xml_dtc_request_register_cookie(request, name, value);
      }
      free(cook);
    }
  }
  return 0;
}

int xml_dtc_request_set_out_header(xml_dtc_request *request, const char *name, const char *value) {
  if (request) {
    if (request->out_headers == NULL)
      request->out_headers = xml_dtc_string_hash_new(request->resource);
    xml_dtc_string_hash_add(request->out_headers, name, value);
    return 0;
  }
  return -1;
}

int xml_dtc_request_apply(xml_dtc_request *request, xml_dtc_list *list) {
  int i, j;

  if (request && list)
    for (i = 0; (request->error.type == NULL) && (request->result == NULL) && (i < list->len); i++) {
      xml_dtc_caller *caller = list->list[i];
      if (!caller->resolved) {
        char *name = caller->component.name;
        caller->component.object = xml_dtc_component_link(xml_dtc_sitemap_get(request->sitemap, caller->name, name));
        free(name);
        caller->resolved = 1;
      }
      if (xml_dtc.hooks)
        for (j = 0; j < xml_dtc.hooks->len; j++) {
          ((xml_dtc_hook)xml_dtc.hooks->list[j])(caller, request);
        }
      if (caller->resolved == 1)
        caller->component.object->handle(caller, request);
      else if (caller->resolved == 2)
        caller->component.handle(caller, request);
    }

  return 0;
}

int xml_dtc_request_set_last_modified(xml_dtc_request *request, time_t last_modified) {
  if (! request)
    return -1;
  
  if (request->last_modified < last_modified)
    request->last_modified = last_modified;

  return 0;
}
