/*
 * This module provides aggregation to the XML-DTC framework
 * Copyright (C) 2003-2004 Mike Hommey <mh@glandium.org>
 * 
 * This module is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This module is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this module; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 */

#include <modules.h>
#include <xml-dtc.h>
#include <mappings.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>
#include <string.h>

typedef struct _xml_dtc_aggregator_caller_data xml_dtc_aggregator_caller_data;

typedef struct _xml_dtc_aggregate_part xml_dtc_aggregate_part;

struct _xml_dtc_aggregator_caller_data {
  xmlChar *element, *ns, *prefix;
  xml_dtc_list *parts;
};

struct _xml_dtc_aggregate_part {
  xmlChar *element, *ns, *prefix, *src;
  int strip_root;
};

static int _xml_dtc_aggregator_caller_free(xml_dtc_aggregator_caller_data *data) {
  if (data) {
    if (data->element)
      xmlFree(data->element);
    if (data->ns)
      xmlFree(data->ns);
    if (data->prefix)
      xmlFree(data->prefix);
    if (data->parts)
      xml_dtc_list_free(data->parts);
    free(data);
  }
  return 0;
}

static int _xml_dtc_aggregate_part_free(xml_dtc_aggregate_part *part) {
  if (part) {
    if (part->element)
      xmlFree(part->element);
    if (part->ns)
      xmlFree(part->ns);
    if (part->prefix)
      xmlFree(part->prefix);
    if (part->src)
      xmlFree(part->src);
    free(part);
  }
  return 0;
}

static int xml_dtc_aggregate(xml_dtc_caller *caller, xml_dtc_request *request) {
  xmlDocPtr xmldoc;
  xmlNodePtr root_node;
  xmlNsPtr ns = NULL;
  xml_dtc_aggregator_caller_data *data = caller->_private;
  int i;

  xmldoc = xmlNewDoc("1.0");

  if (xmldoc == NULL) {
    xml_dtc_msg(caller->resource, XML_DTC_MSG_ERROR, "Could not create XML.");
    return -1;
  }

  if (data->ns)
    ns = xmlNewNs(NULL, data->ns, data->prefix);

  root_node = xmlNewDocNode(xmldoc, ns, data->element, NULL);
  xmlDocSetRootElement(xmldoc, root_node);
  root_node->nsDef=ns;

  for (i = 0; i < data->parts->len; i++) {
    xmlNodePtr node, partroot;
    xmlDocPtr partdoc;
    xml_dtc_aggregate_part *part;
    char *filename;
    part = (xml_dtc_aggregate_part *)data->parts->list[i];
    filename = xml_dtc_resolve_mappings(request->mappings, part->src);
    xml_dtc_sitemap_resolve_relative_path(request->sitemap, &filename);
    if (filename) {
      partdoc = xmlParseFile(filename);
      free(filename);
      partroot = xmlDocGetRootElement(partdoc);
      if (part->strip_root) {
        partroot = partroot->children;
        while (partroot) {
          node = xmlNewChild(root_node, NULL, BAD_CAST "null", NULL);
          xmlReplaceNode(node, xmlDocCopyNode(partroot, xmldoc, 1));
          xmlFreeNode(node);
          partroot = partroot->next;
        }
      } else {
        node = xmlNewChild(root_node, NULL, BAD_CAST "null", NULL);
        xmlReplaceNode(node, xmlDocCopyNode(partroot, xmldoc, 1));
        xmlFreeNode(node);
      }
    /*    if (part->ns)
	  ns = xmlNewNs(NULL, part->ns, part->prefix);*/
      xmlFreeDoc(partdoc);
    }
  }

  request->xmltree = xmldoc;

  return 0;
}

static xml_dtc_component *xml_dtc_handler_aggregate_callback(xml_dtc_handler *handler, xml_dtc_hash *handlers, xml_dtc_sitemap *sitemap, xmlNodePtr node) {
  xml_dtc_caller *result;
  xml_dtc_aggregator_caller_data *data;
  result = xml_dtc_caller_new(sitemap->resource, "aggregator", "aggregator");
  result->resolved = 2;
  result->component.handle = xml_dtc_aggregate;
  result->free_private = (int (*) (void *)) _xml_dtc_aggregator_caller_free;
  data = malloc(sizeof(xml_dtc_aggregator_caller_data));
  memset(data, 0, sizeof(xml_dtc_aggregator_caller_data));
  result->_private = data;
  data->element = xmlGetProp(node, "element");
  data->ns = xmlGetProp(node, "ns");
  data->prefix = xmlGetProp(node, "prefix");

  data->parts = xml_dtc_list_new_cast(sitemap->resource, _xml_dtc_aggregate_part_free, NULL);

  node = node->children;

  while(node) {
    xml_dtc_aggregate_part *part;
    xmlChar *tmp;

    part = malloc(sizeof(xml_dtc_aggregate_part));

    part->src = xmlGetProp(node, "src");
    part->element = xmlGetProp(node, "element");
    part->ns = xmlGetProp(node, "ns");
    part->prefix = xmlGetProp(node, "prefix");
    part->strip_root = xmlStrEqual(tmp = xmlGetProp(node, "strip-root"), "true");
    xmlFree(tmp);

    xml_dtc_list_push(data->parts, part);

    node = node->next;
  }

  return (xml_dtc_component *)result;
}

static int xml_dtc_aggregate_module_register(void) {
  return 0;
}

static const xml_dtc_caller_def xml_dtc_aggregator_caller[] = {
  { "aggregate", "aggregator", (xml_dtc_handler_callback) xml_dtc_handler_aggregate_callback },
  { NULL }
};

xml_dtc_module xml_dtc_aggregate_module = {
  XML_DTC_MODULE_STANDARD_STUFF,
  xml_dtc_aggregate_module_register,
  NULL,
  xml_dtc_aggregator_caller
};
